@@ -28,7 +28,14 @@ from .types import (
     MutationErrorType,
     UploadType,
 )
-from .utils import get_nodes
+from .perms import (
+    check_perms,
+    check_authenticated,
+)
+from .utils import (
+    get_node,
+    get_nodes,
+)
 
 _registry = get_global_registry()
 
@@ -73,7 +80,7 @@ def _get_validation_errors(validation_error):
 def _get_fields(model, only_fields, exclude_fields, required_fields):
     fields = [
         (field.name, field)
-        for field in sorted(list(model._meta.fields))
+        for field in sorted(list(model._meta.fields + model._meta.many_to_many))
     ]
 
     ret = collections.OrderedDict()
@@ -88,11 +95,27 @@ def _get_fields(model, only_fields, exclude_fields, required_fields):
             ret[name] = graphene.ID(
                 description="The ID of the object.",
             )
+        elif isinstance(field, models.FileField):
+            ret[name] = UploadType(
+                description=field.help_text,
+                required=not field.null,
+            )
+        elif isinstance(field, models.BooleanField):
+            ret[name] = graphene.Boolean(
+                description=field.help_text,
+                required=not field.null and not field.blank,
+            )
         elif isinstance(field, (models.ForeignKey, models.OneToOneField)):
             ret[name] = graphene.ID(
                 description=field.help_text,
                 required=not field.null,
             )
+        elif isinstance(field, models.ManyToManyField):
+            ret[name] = graphene.List(
+                graphene.ID,
+                description=field.help_text,
+                required=not field.null,
+            )
         else:
             ret[name] = convert_django_field_with_choices(field, _registry)
 
@@ -128,6 +151,9 @@ class BaseMutationOptions(MutationOptions):
     #: A list of Django permissions to check against the user
     permissions = None
 
+    #: If any permission should allow the user to execute this mutation
+    permissions_any = None
+
     #: If we should allow unauthenticated users to do this mutation
     allow_unauthenticated = False
 
@@ -146,12 +172,14 @@ class BaseMutation(ClientIDMutation):
 
     @classmethod
     def __init_subclass_with_meta__(cls, permissions=None,
+                                    permissions_any=True,
                                     allow_unauthenticated=False,
                                     _meta=None, **kwargs):
         if not _meta:
             _meta = BaseMutationOptions(cls)
 
         _meta.permissions = permissions or []
+        _meta.permissions_any = permissions_any
         _meta.allow_unauthenticated = allow_unauthenticated
 
         super().__init_subclass_with_meta__(_meta=_meta, **kwargs)
@@ -163,11 +191,7 @@ class BaseMutation(ClientIDMutation):
             return None
 
         try:
-            node = graphene.Node.get_node_from_global_id(
-                info,
-                node_id,
-                only_type,
-            )
+            node = get_node(node_id, only_type)
         except (AssertionError, GraphQLError) as e:
             raise ValidationError({field: str(e)})
         else:
@@ -195,13 +219,14 @@ class BaseMutation(ClientIDMutation):
         Subclasses can override this to avoid the permission checking or
         extending it. Remember to call `super()` in the later case.
         """
-        if not cls._meta.allow_unauthenticated and not user.is_authenticated:
+        if not cls._meta.allow_unauthenticated and not check_authenticated(user):
             return False
 
         if not cls._meta.permissions:
             return True
 
-        return user.has_perms(cls._meta.permissions)
+        return check_perms(user, cls._meta.permissions,
+                           any_perm=cls._meta.permissions_any)
 
     @classmethod
     def mutate_and_get_payload(cls, root, info, **data):
