@@ -1,9 +1,7 @@
-import types
-
 from django.db import models
+from django.db.models import Prefetch
 import graphene
 from graphene_django import DjangoObjectType
-from graphene_django.fields import DjangoConnectionField
 from graphene_django.types import DjangoObjectTypeOptions
 
 try:
@@ -14,6 +12,10 @@ except ImportError:
     _BaseDjangoObjectType = DjangoObjectType
 
 from .exceptions import PermissionDenied
+from .perms import (
+    check_perms,
+    check_authenticated,
+)
 from .models import (
     GuardedModel,
     GuardedModelManager,
@@ -68,6 +70,9 @@ class ModelTypeOptions(DjangoObjectTypeOptions):
     #: query this model.
     permissions = None
 
+    #: If any permission should allow the user to query this model.
+    permissions_any = True
+
     #: A list of guardian object permissions to check if the user has
     #: permission to query the model object.
     object_permissions = None
@@ -75,12 +80,6 @@ class ModelTypeOptions(DjangoObjectTypeOptions):
     #: If any object permission should allow the user to query this model.
     object_permissions_any = True
 
-    #: A dict defining fields to be prefetched. This should map the
-    #: related name to the type. e.g. `{'items': ItemType}`.
-    #: Note that for this to work, `graphene_django_optimizer` needs
-    #: to be installed
-    prefetch = None
-
 
 class ModelType(_BaseDjangoObjectType):
     """Base type with automatic optimizations and permissions checking."""
@@ -91,6 +90,7 @@ class ModelType(_BaseDjangoObjectType):
     @classmethod
     def __init_subclass_with_meta__(cls, _meta=None, model=None,
                                     permissions=None,
+                                    permissions_any=True,
                                     object_permissions=None,
                                     object_permissions_any=True,
                                     allow_unauthenticated=False,
@@ -100,24 +100,11 @@ class ModelType(_BaseDjangoObjectType):
             _meta = DjangoObjectTypeOptions(cls)
 
         _meta.permissions = permissions or []
+        _meta.permissions_any = permissions_any
         _meta.object_permissions = object_permissions or []
         _meta.object_permissions_any = object_permissions_any
         _meta.allow_unauthenticated = allow_unauthenticated
 
-        # Automatic prefetch optimization
-        if gql_optimizer is not None:
-            prefetch = prefetch or {}
-            for k, v in prefetch.items():
-                if isinstance(v, types.FunctionType):
-                    v = v()
-                f = gql_optimizer.field(
-                    DjangoConnectionField(v),
-                    model_field=k,
-                )
-                r = lambda s, i, **kw: getattr(s, k).all()
-                setattr(cls, k, f)
-                setattr(cls, 'resolve_' + k, r)
-
         super().__init_subclass_with_meta__(_meta=_meta, model=model, **kwargs)
 
     @classmethod
@@ -131,12 +118,13 @@ class ModelType(_BaseDjangoObjectType):
         `graphene_django_optimizer` is installed.
         """
         if not cls.check_permissions(info.context.user):
-            raise PermissionDenied()
+            raise PermissionDenied("No permissions")
+
+        if isinstance(qs, models.Manager):
+            qs = qs.get_queryset()
 
         if (cls._meta.object_permissions and
                 isinstance(cls._meta.model.objects, GuardedModelManager)):
-            if isinstance(qs, models.Manager):
-                qs = qs.get_queryset()
             qs &= cls._meta.model.objects.for_user(
                 info.context.user,
                 cls._meta.object_permissions,
@@ -147,15 +135,29 @@ class ModelType(_BaseDjangoObjectType):
         if gql_optimizer is None:
             return ret
 
-        return gql_optimizer.query(ret, info)
+        ret = gql_optimizer.query(ret, info)
+        prl = {i.to_attr if isinstance(i, Prefetch) else i: i
+               for i in ret._prefetch_related_lookups}
+        ret._prefetch_related_lookups = tuple(prl.values())
+
+        return ret
 
     @classmethod
     def get_node(cls, info, id):
         """Get the node instance given the relay global id."""
-        instance = super().get_node(info, id)
+        # NOTE: get_queryset will filter allowed models for the user so
+        # this will return None if he is not allowed to retrieve this
 
-        if not cls.check_object_permissions(info.context.user, instance):
-            raise PermissionDenied()
+        if gql_optimizer is not None:
+            # optimizer will ignore queryset so call the same as they call
+            # but passing objects from get_queryset to keep our preferences
+            instance = cls.maybe_optimize(
+                info,
+                cls.get_queryset(cls._meta.model.objects, info),
+                id,
+            )
+        else:
+            instance = super().get_node(info, id)
 
         return instance
 
@@ -166,13 +168,14 @@ class ModelType(_BaseDjangoObjectType):
         Subclasses can override this to avoid the permission checking or
         extending it. Remember to call `super()` in the later case.
         """
-        if not cls._meta.allow_unauthenticated and not user.is_authenticated:
+        if not cls._meta.allow_unauthenticated and not check_authenticated(user):
             return False
 
         if not cls._meta.permissions:
             return True
 
-        return user.has_perms(cls._meta.permissions)
+        return check_perms(user, cls._meta.permissions,
+                           any_perm=cls._meta.permissions_any)
 
     @classmethod
     def check_object_permissions(cls, user, instance):
